"""
Retrieving user information from an LDAP directory.

"""
import collections
import dataclasses
import typing

import ldap3

from .config import ConfigurationDataclassMixin


# User information we need to populate the Google user directory.
UserEntry = collections.namedtuple('UserEntry', 'uid cn sn displayName')


@dataclasses.dataclass
class Configuration(ConfigurationDataclassMixin):
    """
    Configuration for accessing the LDAP directory.

    """
    host: str

    search_base: str

    eligible_user_filter: str

    managed_user_filter: typing.Union[str, None] = None

    def get_eligible_uids(self):
        """
        Return a set containing all uids who are eligible to have a Google account.

        """
        return {
            e['attributes']['uid'][0]
            for e in self._search(search_filter=self.eligible_user_filter, attributes=['uid'])
        }

    def get_managed_user_entries(self):
        """
        Return a list containing all managed user entries as UserEntry instances.

        """
        search_filter = (
            self.managed_user_filter
            if self.managed_user_filter is not None
            else self.eligible_user_filter
        )
        return [
            UserEntry(
                uid=_extract(e, 'uid'), cn=_extract(e, 'cn'), sn=_extract(e, 'sn'),
                displayName=_extract(e, 'displayName')
            )
            for e in self._search(
                search_filter=search_filter, attributes=['uid', 'cn', 'sn', 'displayName']
            )
        ]

    def _search(self, *, search_filter, attributes):
        ldap_server = ldap3.Server(self.host)
        with ldap3.Connection(ldap_server, auto_bind=True) as conn:
            return conn.extend.standard.paged_search(
                self.search_base, search_filter, paged_size=1000, attributes=attributes)


def _extract(entry, attr, *, default=''):
    vs = entry['attributes'].get(attr, [])
    if len(vs) == 0:
        return default
    if isinstance(vs, str):
        return vs
    return vs[0]
