"""
Synchronisation limits.

"""
import dataclasses
import numbers
import typing

from . import config


@dataclasses.dataclass
class Configuration(config.ConfigurationDataclassMixin):
    """
    Configuration for synchronisation limits.

    """
    # The abort_... settings below are safety limits and will abort the run if the limits are
    # violated. They are there to define the "sane limits" for an update.

    # Refuse to perform sync if we are to "touch" more than this percentage of users. The
    # percentage of users "touched" is calculated as
    #
    #   (new google users + modified google users) / max(1, total google users)
    #
    # where "modified" includes metadata changes and suspension/restoration. As such this
    # calculated percentage can be greater than 100. Set to null to have no limit. Default: null.
    abort_user_change_percentage: typing.Union[None, numbers.Real] = None

    # The max_... settings below will not abort the run if the number of users affected is greater
    # than the specified number. Instead the number of users affected is capped to that number. The
    # selection of which users are included in the capped number is arbitrary.

    # Limit the number of new user creations per run. This is an absolute number. Set to None to
    # have no limit.
    max_new_users: typing.Union[None, numbers.Real] = None

    # Limit the number of user suspensions per run. This is an absolute number. Set to None to
    # have no limit.
    max_suspended_users: typing.Union[None, numbers.Real] = None

    # Limit the number of user un-suspensions (reactivations) per run. This is an absolute number.
    # Set to None to have no limit.
    max_reactivated_users: typing.Union[None, numbers.Real] = None

    # Limit the number of user metadata changes per run. This is an absolute number. Set to None to
    # have no limit.
    max_updated_users: typing.Union[None, numbers.Real] = None
